"""Provide a data model for domains and some utility functions."""

# Standard Python Libraries
from collections import OrderedDict
from datetime import datetime, timedelta
from os import path, stat
from typing import Dict

# Third-Party Libraries
import publicsuffix

from . import PublicSuffixListFilename, PublicSuffixListReadOnly, trustymail


def get_psl():
    """
    Get the Public Suffix List - either new, or cached in the CWD for 24 hours.

    Returns
    -------
    PublicSuffixList: An instance of PublicSuffixList loaded with a cached or updated list
    """

    def download_psl():
        fresh_psl = publicsuffix.fetch()
        with open(PublicSuffixListFilename, "w", encoding="utf-8") as fresh_psl_file:
            fresh_psl_file.write(fresh_psl.read())

    # Download the psl if necessary
    if not PublicSuffixListReadOnly:
        if not path.exists(PublicSuffixListFilename):
            download_psl()
        else:
            psl_age = datetime.now() - datetime.fromtimestamp(
                stat(PublicSuffixListFilename).st_mtime
            )
            if psl_age > timedelta(hours=24):
                download_psl()

    with open(PublicSuffixListFilename, encoding="utf-8") as psl_file:
        psl = publicsuffix.PublicSuffixList(psl_file)

    return psl


def get_public_suffix(domain):
    """Return the public suffix of a given domain."""
    public_list = get_psl()

    return public_list.get_public_suffix(domain)


def format_list(record_list):
    """Format a list into a string to increase readability in CSV."""
    # record_list should only be a list, not an integer, None, or
    # anything else.  Thus this if clause handles only empty
    # lists.  This makes a "null" appear in the JSON output for
    # empty lists, as expected.
    if not record_list:
        return None

    return ", ".join(record_list)


class Domain:
    """Store information about a domain."""

    base_domains: Dict[str, "Domain"] = {}

    def __init__(
        self,
        domain_name,
        timeout,
        smtp_timeout,
        smtp_localhost,
        smtp_ports,
        smtp_cache,
        dns_hostnames,
    ):
        """Retrieve information about a given domain name."""
        self.domain_name = domain_name.lower()

        self.base_domain_name = get_public_suffix(self.domain_name)

        self.is_base_domain = True
        self.base_domain = None
        if self.base_domain_name != self.domain_name:
            self.is_base_domain = False
            if self.base_domain_name not in Domain.base_domains:
                # Populate DMARC for parent.
                domain = trustymail.scan(
                    self.base_domain_name,
                    timeout,
                    smtp_timeout,
                    smtp_localhost,
                    smtp_ports,
                    smtp_cache,
                    {"mx": False, "starttls": False, "spf": False, "dmarc": True},
                    dns_hostnames,
                )
                Domain.base_domains[self.base_domain_name] = domain
            self.base_domain = Domain.base_domains[self.base_domain_name]

        # Start off assuming the host is live unless an error tells us otherwise.
        self.is_live = True

        # Keep entire record for potential future use.
        self.mx_records = None
        self.mx_records_dnssec = None
        self.spf = None
        self.spf_dnssec = None
        self.dmarc = None
        self.dmarc_dnssec = False
        self.dmarc_policy = None
        self.dmarc_subdomain_policy = None
        self.dmarc_pct = None
        self.dmarc_aggregate_uris = []
        self.dmarc_forensic_uris = []
        self.dmarc_has_aggregate_uri = False
        self.dmarc_has_forensic_uri = False
        self.dmarc_reports_address_error = False

        # Syntax validity - default spf to false as the lack of an SPF is a bad thing.
        self.valid_spf = False
        self.valid_dmarc = True
        self.syntax_errors = []

        # Mail Info
        self.mail_servers = None

        # A dictionary for each port for each entry in mail_servers.
        # The dictionary's values indicate:
        # 1. Whether or not the server is listening on the port
        # 2. Whether or not the server supports SMTP
        # 3. Whether or not the server supports STARTTLS
        self.starttls_results = {}

        # A list of any debugging information collected while scanning records.
        self.debug_info = []

        # A list of the ports tested for SMTP
        self.ports_tested = set()

    def has_mail(self):
        """Check if there are any mail servers associated with this domain."""
        if self.mail_servers is not None:
            return len(self.mail_servers) > 0
        return None

    def has_supports_smtp(self):
        """Check if any of the mail servers associated with this domain are listening and support SMTP."""
        result = None
        if len(self.starttls_results) > 0:
            result = (
                len(
                    filter(
                        lambda x: self.starttls_results[x]["supports_smtp"],
                        self.starttls_results.keys(),
                    )
                )
                > 0
            )
        return result

    def has_starttls(self):
        """Check if any of the mail servers associated with this domain are listening and support STARTTLS."""
        result = None
        if len(self.starttls_results) > 0:
            result = (
                len(
                    filter(
                        lambda x: self.starttls_results[x]["starttls"],
                        self.starttls_results.keys(),
                    )
                )
                > 0
            )
        return result

    def has_spf(self):
        """Check if this domain has any Sender Policy Framework records."""
        if self.spf is not None:
            return len(self.spf) > 0
        return None

    def has_dmarc(self):
        """Check if this domain has a Domain-based Message Authentication, Reporting, and Conformance record."""
        if self.dmarc is not None:
            return len(self.dmarc) > 0
        return None

    def add_mx_record(self, record):
        """Add a mail server record for this domain."""
        if self.mx_records is None:
            self.mx_records = []
        self.mx_records.append(record)
        # The rstrip is because dnspython's string representation of
        # the record will contain a trailing period if it is a FQDN.
        if self.mail_servers is None:
            self.mail_servers = []
        self.mail_servers.append(record.exchange.to_text().rstrip(".").lower())

    def parent_has_dmarc(self):
        """Check if a domain or its parent has a Domain-based Message Authentication, Reporting, and Conformance record."""
        ans = self.has_dmarc()
        if self.base_domain:
            ans = self.base_domain.has_dmarc()
        return ans

    def parent_dmarc_dnssec(self):
        """Get this domain or its parent's DMARC DNSSEC information."""
        ans = self.dmarc_dnssec
        if self.base_domain:
            ans = self.base_domain.dmarc_dnssec
        return ans

    def parent_valid_dmarc(self):
        """Check if this domain or its parent have a valid DMARC record."""
        ans = self.valid_dmarc
        if self.base_domain:
            return self.base_domain.valid_dmarc
        return ans

    def parent_dmarc_results(self):
        """Get this domain or its parent's DMARC information."""
        ans = format_list(self.dmarc)
        if self.base_domain:
            ans = format_list(self.base_domain.dmarc)
        return ans

    def get_dmarc_policy(self):
        """Get this domain or its parent's DMARC policy."""
        ans = self.dmarc_policy
        # If the policy was never set, or isn't in the list of valid
        # policies, check the parents.
        if ans is None or ans.lower() not in ["quarantine", "reject", "none"]:
            if self.base_domain:
                # We check the *subdomain* policy in case one was
                # explicitly set.  If one was not explicitly set then
                # the subdomain policy is populated with the value for
                # the domain policy by trustymail.py anyway, in
                # accordance with the RFC
                # (https://tools.ietf.org/html/rfc7489#section-6.3).
                ans = self.base_domain.get_dmarc_subdomain_policy()
            else:
                ans = None
        return ans

    def get_dmarc_subdomain_policy(self):
        """Get this domain or its parent's DMARC subdomain policy."""
        ans = self.dmarc_subdomain_policy
        # If the policy was never set, or isn't in the list of valid
        # policies, check the parents.
        if ans is None or ans.lower() not in ["quarantine", "reject", "none"]:
            if self.base_domain:
                ans = self.base_domain.get_dmarc_subdomain_policy()
            else:
                ans = None
        return ans

    def get_dmarc_pct(self):
        """Get this domain or its parent's DMARC percentage information."""
        ans = self.dmarc_pct
        if not ans and self.base_domain:
            # Check the parents
            ans = self.base_domain.get_dmarc_pct()
        return ans

    def get_dmarc_has_aggregate_uri(self):
        """Get this domain or its parent's DMARC aggregate URI."""
        ans = self.dmarc_has_aggregate_uri
        # If there are no aggregate URIs then check the parents.
        if not ans and self.base_domain:
            ans = self.base_domain.get_dmarc_has_aggregate_uri()
        return ans

    def get_dmarc_has_forensic_uri(self):
        """Check if this domain or its parent have a DMARC forensic URI."""
        ans = self.dmarc_has_forensic_uri
        # If there are no forensic URIs then check the parents.
        if not ans and self.base_domain:
            ans = self.base_domain.get_dmarc_has_forensic_uri()
        return ans

    def get_dmarc_aggregate_uris(self):
        """Get this domain or its parent's DMARC aggregate URIs."""
        ans = self.dmarc_aggregate_uris
        # If there are no aggregate URIs then check the parents.
        if not ans and self.base_domain:
            ans = self.base_domain.get_dmarc_aggregate_uris()
        return ans

    def get_dmarc_forensic_uris(self):
        """Get this domain or its parent's DMARC forensic URIs."""
        ans = self.dmarc_forensic_uris
        # If there are no forensic URIs then check the parents.
        if not ans and self.base_domain:
            ans = self.base_domain.get_dmarc_forensic_uris()
        return ans

    def generate_results(self):
        """Generate the results for this domain."""
        if len(self.starttls_results.keys()) == 0:
            domain_supports_smtp = None
            domain_supports_starttls = None
            mail_servers_that_support_smtp = None
            mail_servers_that_support_starttls = None
        else:
            mail_servers_that_support_smtp = [
                x
                for x in self.starttls_results.keys()
                if self.starttls_results[x]["supports_smtp"]
            ]
            mail_servers_that_support_starttls = [
                x
                for x in self.starttls_results.keys()
                if self.starttls_results[x]["starttls"]
            ]
            domain_supports_smtp = bool(mail_servers_that_support_smtp)
            domain_supports_starttls = domain_supports_smtp and all(
                [
                    self.starttls_results[x]["starttls"]
                    for x in mail_servers_that_support_smtp
                ]
            )

        results = OrderedDict(
            [
                ("Domain", self.domain_name),
                ("Base Domain", self.base_domain_name),
                ("Live", self.is_live),
                ("MX Record", self.has_mail()),
                ("MX Record DNSSEC", self.mx_records_dnssec),
                ("Mail Servers", format_list(self.mail_servers)),
                (
                    "Mail Server Ports Tested",
                    format_list([str(port) for port in self.ports_tested]),
                ),
                (
                    "Domain Supports SMTP Results",
                    format_list(mail_servers_that_support_smtp),
                ),
                # True if and only if at least one mail server speaks SMTP
                ("Domain Supports SMTP", domain_supports_smtp),
                (
                    "Domain Supports STARTTLS Results",
                    format_list(mail_servers_that_support_starttls),
                ),
                # True if and only if all mail servers that speak SMTP
                # also support STARTTLS
                ("Domain Supports STARTTLS", domain_supports_starttls),
                ("SPF Record", self.has_spf()),
                ("SPF Record DNSSEC", self.spf_dnssec),
                ("Valid SPF", self.valid_spf),
                ("SPF Results", format_list(self.spf)),
                ("DMARC Record", self.has_dmarc()),
                ("DMARC Record DNSSEC", self.dmarc_dnssec),
                ("Valid DMARC", self.has_dmarc() and self.valid_dmarc),
                ("DMARC Results", format_list(self.dmarc)),
                ("DMARC Record on Base Domain", self.parent_has_dmarc()),
                ("DMARC Record on Base Domain DNSSEC", self.parent_dmarc_dnssec()),
                (
                    "Valid DMARC Record on Base Domain",
                    self.parent_has_dmarc() and self.parent_valid_dmarc(),
                ),
                ("DMARC Results on Base Domain", self.parent_dmarc_results()),
                ("DMARC Policy", self.get_dmarc_policy()),
                ("DMARC Subdomain Policy", self.get_dmarc_subdomain_policy()),
                ("DMARC Policy Percentage", self.get_dmarc_pct()),
                (
                    "DMARC Aggregate Report URIs",
                    format_list(self.get_dmarc_aggregate_uris()),
                ),
                (
                    "DMARC Forensic Report URIs",
                    format_list(self.get_dmarc_forensic_uris()),
                ),
                ("DMARC Has Aggregate Report URI", self.get_dmarc_has_aggregate_uri()),
                ("DMARC Has Forensic Report URI", self.get_dmarc_has_forensic_uri()),
                (
                    "DMARC Reporting Address Acceptance Error",
                    self.dmarc_reports_address_error,
                ),
                ("Syntax Errors", format_list(self.syntax_errors)),
                ("Debug Info", format_list(self.debug_info)),
            ]
        )

        return results



"""The trustymail library."""
# Standard Python Libraries
from typing import List

# We disable a Flake8 check for "Module imported but unused (F401)" here because
# although this import is not directly used, it populates the value
# package_name.__version__, which is used to get version information about this
# Python package.
from ._version import __version__  # noqa: F401

__all__: List[str] = []

PublicSuffixListFilename = "public_suffix_list.dat"
PublicSuffixListReadOnly = False



"""trustymail: A tool for scanning DNS mail records for evaluating security.

Usage:
  trustymail (INPUT ...) [options]
  trustymail (INPUT ...) [--output=OUTFILE] [--timeout=TIMEOUT] [--smtp-timeout=TIMEOUT] [--smtp-localhost=HOSTNAME] [--smtp-ports=PORTS] [--no-smtp-cache] [--mx] [--starttls] [--spf] [--dmarc] [--debug] [--json] [--dns=HOSTNAMES] [--psl-filename=FILENAME] [--psl-read-only]
  trustymail (-h | --help)

Options:
  -h --help                   Show this message.
  -o --output=OUTFILE         Name of output file.  (Default results)
  -t --timeout=TIMEOUT        The DNS lookup timeout in seconds.  (Default is 5.)
  --smtp-timeout=TIMEOUT      The SMTP connection timeout in seconds.  (Default is 5.)
  --smtp-localhost=HOSTNAME   The hostname to use when connecting to SMTP
                              servers.  (Default is the FQDN of the host from
                              which trustymail is being run.)
  --smtp-ports=PORTS          A comma-delimited list of ports at which to look
                              for SMTP servers.  (Default is '25,465,587'.)
  --no-smtp-cache             Do not cache SMTP results during the run.  This
                              may results in slower scans due to testing the
                              same mail servers multiple times.
  --mx                        Only check MX records.
  --starttls                  Only check MX records and STARTTLS support.
                              (Implies --mx.)
  --spf                       Only check SPF records.
  --dmarc                     Only check DMARC records.
  --json                      Output is in JSON format.  (Default is CSV.)
  --debug                     Output should include more verbose logging.
  --dns=HOSTNAMES             A comma-delimited list of DNS servers to query
                              against.  For example, if you want to use
                              Google's DNS then you would use the
                              value --dns-hostnames='8.8.8.8,8.8.4.4'.  By
                              default the DNS configuration of the host OS
                              (/etc/resolv.conf) is used.  Note that
                              the host's DNS configuration is not used at all
                              if this option is used.
  --psl-filename=FILENAME     The name of the file where the public suffix list
                              (PSL) cache will be saved.  If set to the name of
                              an existing file then that file will be used as
                              the PSL.  If not present then the PSL cache will
                              be saved to a file in the current directory called
                              public_suffix_list.dat.
  --psl-read-only             If present, then the public suffix list (PSL)
                              cache will be read but never overwritten.  This
                              is useful when running in AWS Lambda, for
                              instance, where the local filesystem is read-only.

Notes:
   If no scan type options are specified, all are run against a given domain/input.
"""
# Standard Python Libraries
# Built-in imports
import errno
import logging
import os

# Third-Party Libraries
# Dependency imports
import docopt

# Local imports
from . import trustymail
from ._version import __version__

# The default ports to be checked to see if an SMTP server is listening.
_DEFAULT_SMTP_PORTS = {25, 465, 587}


def main():
    """Perform a trustymail scan using the provided options."""
    args = docopt.docopt(__doc__, version=__version__)

    # Monkey patching trustymail to make it cache the PSL where we want
    if args["--psl-filename"] is not None:
        trustymail.PublicSuffixListFilename = args["--psl-filename"]
    # Monkey patching trustymail to make the PSL cache read-only
    if args["--psl-read-only"]:
        trustymail.PublicSuffixListReadOnly = True
    # cisagov Libraries
    import trustymail.trustymail as tmail

    log_level = logging.WARN
    if args["--debug"]:
        log_level = logging.DEBUG
    logging.basicConfig(format="%(asctime)-15s %(message)s", level=log_level)

    # Allow for user to input a csv for many domain names.
    if args["INPUT"][0].endswith(".csv"):
        domains = tmail.domain_list_from_csv(open(args["INPUT"][0]))
    else:
        domains = args["INPUT"]

    if args["--timeout"] is not None:
        timeout = int(args["--timeout"])
    else:
        timeout = 5

    if args["--smtp-timeout"] is not None:
        smtp_timeout = int(args["--smtp-timeout"])
    else:
        smtp_timeout = 5

    if args["--smtp-localhost"] is not None:
        smtp_localhost = args["--smtp-localhost"]
    else:
        smtp_localhost = None

    if args["--smtp-ports"] is not None:
        smtp_ports = {int(port) for port in args["--smtp-ports"].split(",")}
    else:
        smtp_ports = _DEFAULT_SMTP_PORTS

    if args["--dns"] is not None:
        dns_hostnames = args["--dns"].split(",")
    else:
        dns_hostnames = None

    # --starttls implies --mx
    if args["--starttls"]:
        args["--mx"] = True

    # User might not want every scan performed.
    scan_types = {
        "mx": args["--mx"],
        "starttls": args["--starttls"],
        "spf": args["--spf"],
        "dmarc": args["--dmarc"],
    }

    domain_scans = []
    for domain_name in domains:
        domain_scans.append(
            tmail.scan(
                domain_name,
                timeout,
                smtp_timeout,
                smtp_localhost,
                smtp_ports,
                not args["--no-smtp-cache"],
                scan_types,
                dns_hostnames,
            )
        )

    # Default output file name is results.
    if args["--output"] is None:
        output_file_name = "results"
    else:
        output_file_name = args["--output"]

    # Ensure file extension is present in filename.
    if args["--json"] and ".json" not in output_file_name:
        output_file_name += ".json"
    elif ".csv" not in output_file_name:
        output_file_name += ".csv"

    if args["--json"]:
        json_out = tmail.generate_json(domain_scans)
        if args["--output"] is None:
            print(json_out)
        else:
            write(json_out, output_file_name)
            logging.warn("Wrote results to %s." % output_file_name)
    else:
        tmail.generate_csv(domain_scans, output_file_name)


def write(content, out_file):
    """Write the provided content to a file after ensuring all intermediate directories exist."""
    parent = os.path.dirname(out_file)
    if parent != "":
        mkdir_p(parent)

    f = open(out_file, "w")  # no utf-8 in python 2
    f.write(content)
    f.close()


# mkdir -p in python, from:
# http://stackoverflow.com/questions/600268/mkdir-p-functionality-in-python
def mkdir_p(path):
    """Make a directory and all intermediate directories in its path."""
    try:
        os.makedirs(path)
    except OSError as exc:  # Python >2.5
        if exc.errno == errno.EEXIST:
            pass
        else:
            raise



"""
This is the setup module for the trustymail project.

Based on:

- https://packaging.python.org/distributing/
- https://github.com/pypa/sampleproject/blob/master/setup.py
- https://blog.ionelmc.ro/2014/05/25/python-packaging/#the-structure
"""

# Standard Python Libraries
import codecs
from glob import glob
from os.path import abspath, basename, dirname, join, splitext

# Third-Party Libraries
from setuptools import find_packages, setup


def readme():
    """Read in and return the contents of the project's README.md file."""
    with open("README.md", encoding="utf-8") as f:
        return f.read()


# Below two methods were pulled from:
# https://packaging.python.org/guides/single-sourcing-package-version/
def read(rel_path):
    """Open a file for reading from a given relative path."""
    here = abspath(dirname(__file__))
    with codecs.open(join(here, rel_path), "r") as fp:
        return fp.read()


def get_version(version_file):
    """Extract a version number from the given file path."""
    for line in read(version_file).splitlines():
        if line.startswith("__version__"):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    raise RuntimeError("Unable to find version string.")


setup(
    name="trustymail",
    # Versions should comply with PEP440
    version=get_version("src/trustymail/_version.py"),
    description="Scan domains and return data based on trustworthy email best practices",
    long_description=readme(),
    long_description_content_type="text/markdown",
    # Landing page for CISA's cybersecurity mission
    url="https://www.cisa.gov/cybersecurity",
    # Additional URLs for this project per
    # https://packaging.python.org/guides/distributing-packages-using-setuptools/#project-urls
    project_urls={
        "Source": "https://github.com/cisagov/trustymail",
        "Tracker": "https://github.com/cisagov/trustymail/issues",
    },
    # Author details
    author="Cybersecurity and Infrastructure Security Agency",
    author_email="github@cisa.dhs.gov",
    license="License :: CC0 1.0 Universal (CC0 1.0) Public Domain Dedication",
    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        "Development Status :: 4 - Beta",
        # Indicate who your project is intended for
        "Intended Audience :: Developers",
        # Pick your license as you wish (should match "license" above)
        "License :: CC0 1.0 Universal (CC0 1.0) Public Domain Dedication",
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: Implementation :: CPython",
    ],
    python_requires=">=3.6",
    # What does your project relate to?
    keywords="email authentication STARTTLS",
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    py_modules=[splitext(basename(path))[0] for path in glob("src/*.py")],
    install_requires=[
        "dnspython",
        "docopt",
        "publicsuffix",
        "py3dns",
        "pyspf",
        "requests",
        "setuptools >= 24.2.0",
    ],
    extras_require={
        "test": [
            "coverage",
            # coveralls 1.11.0 added a service number for calls from
            # GitHub Actions. This caused a regression which resulted in a 422
            # response from the coveralls API with the message:
            # Unprocessable Entity for url: https://coveralls.io/api/v1/jobs
            # 1.11.1 fixed this issue, but to ensure expected behavior we'll pin
            # to never grab the regression version.
            "coveralls != 1.11.0",
            "pre-commit",
            "pytest-cov",
            "pytest",
        ]
    },
    # Conveniently allows one to run the CLI tool as `trustymail`
    entry_points={"console_scripts": ["trustymail = trustymail.cli:main"]},
)
